/*	lemon_board Copyright (C) 2022 Wael Karram.
*
*	This program is free software: you can redistribute it and/or modify
*	it under the terms of the GNU Affero General Public License as published by
*	the Free Software Foundation, version 3 of the License.
*
*	This program is distributed in the hope that it will be useful,
*	but WITHOUT ANY WARRANTY; without even the implied warranty of
*	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*	GNU Affero General Public License for more details.
*
*	You should have received a copy of the GNU Affero General Public License
*	along with this program.  If not, see <https://www.gnu.org/licenses/>.
*/

/* Include the header. */
#include "post.h"

/* Serialization function for a post_t struct. */
/* Returns NULL on allocation failure. */
static inline char *serialize_post(post_t *post) {
	/* Sanity check. */
	assert(post);

	/* Local variables. */
	char **strings, *buffer, *tmp_buffer;
	size_t i = 0L;
	size_t length, total_length, i, j;

	/* Allocate the total array of strings. */
	/* Current post, referencing posts, referenced posts, text and path. */
	strings = (char**) calloc(3L + post->referenced_posts_length + post->referencing_posts_length, sizeof(char*));
	if (!strings)
		return NULL;

	/* Start by converting the OP.*/
	strings[i] = ulltostr(post->id);
	/* Add delimeter. */
	if (strings[i]) {
		length = strlen(strings[i]);
		tmp_buffer = realloc(strings[i], (length + 2L) * sizeof(char));
		if (!tmp_buffer) {
			free(strings[i]);
			free(strings);
			return NULL;
		}
		tmp_buffer[length - 1L] = POST_REFERENCED_ID_START_DELIMETER_POST_H;
		tmp_buffer[length] = NULL_CHAR_POST_H;
		/* Move back. */
		strings[i] = tmp_buffer;
 	} else {
		free(strings);
		return NULL;
	}
	/* Add to the total length. */
	total_length = length;

	/* Convert the referenced IDs. */
	for (i = 1L; i <= post->referenced_posts_length; i++) {
		strings[i] = ulltostr(post->referenced_posts[i - 1L]);
		/* Add delimeter. */
		if (strings[i]) {
			length = strlen(strings[i]);
			tmp_buffer = realloc(strings[i], (length + 2L) * sizeof(char));
			if (!tmp_buffer) {
				for (j = 0L; j < i; j++)
					free(strings[j]);
				free(strings);
				return NULL;
			}
			tmp_buffer[length - 1L] = (i != post->referenced_posts_length) ? POST_REFERENCED_ID_DELIMETER_POST_H : POST_REFERENCING_ID_START_DELIMETER_POST_H;
			tmp_buffer[length] = NULL_CHAR_POST_H;
			/* Move back. */
			strings[i] = tmp_buffer;
		} else {
			/* Cleanup. */
			for (j = 0L; j < i; j++)
				free(strings[j]);
			free(strings);
			return NULL;
		}
	}

	/* Add the start delimeter for referenced IDs if there were no referenced posts. */
	if (!post->referenced_posts_length) {
		length = strlen(strings[i]);
		tmp_buffer = realloc(strings[i], (length + 2L) * sizeof(char));
		if (!tmp_buffer) {
			for (j = 0L; j < i; j++)
				free(strings[j]);
			free(strings);
			return NULL;
		}
		tmp_buffer[length - 1L] = POST_REFERENCING_ID_START_DELIMETER_POST_H;
		tmp_buffer[length] = NULL_CHAR_POST_H;
		/* Move back. */
		strings[i] = tmp_buffer;
	}

	/* Convert the referencing IDs. */
	for (; i <= post->referenced_posts_length + post->referencing_posts_length; i++) {
		strings[i] = ulltostr(post->referencing_posts[i - 1L - post->referenced_posts_length]);
		/* Add delimeter. */
		if (strings[i]) {
			length = strlen(strings[i]);
			tmp_buffer = realloc(strings[i], (length + 2L) * sizeof(char));
			if (!tmp_buffer) {
				for (j = 0L; j < i; j++)
					free(strings[i]);
				free(strings);
				return NULL;
			}
			tmp_buffer[length - 1L] = (i != (post->referenced_posts_length + post->refencing_posts_length)) ? POST_REFERENCING_ID_DELIMETER_POST_H : LINE_DELIMETER_POST_H;
			tmp_buffer[length] = NULL_CHAR_POST_H;
			/* Move back. */
			strings[i] = tmp_buffer;
		} else {
			/* Cleanup. */
			for (j = 0L; j < i; j++)
				free(strings[i]);
			free(strings);
			return NULL;
		}
	}

	/* Add the start delimeter for referenced IDs if there were no referenced posts. */
	if (!post->referencing_posts_length) {
		length = strlen(strings[i]);
		tmp_buffer = realloc(strings[i], (length + 2L) * sizeof(char));
		if (!tmp_buffer) {
			for (j = 0L; j < i; j++)
				free(strings[j]);
			free(strings);
			return NULL;
		}
		tmp_buffer[length - 1L] = LINE_DELIMETER_POST_H;
		tmp_buffer[length] = NULL_CHAR_POST_H;
		/* Move back. */
		strings[i] = tmp_buffer;
	}

	/* Add the image path. */
	length = strlen(post->image_path) + strlen(IMAGE_PATH_PREFIX_POST_H) + strlen(NULL_CHAR_POST_H) + 2L;
	strings[i] = (char*) calloc(length, sizeof(char));
	if (!strings[i]) {
		for (j = 0L; j < i; j++)
			free(strings[i]);
		free(strings);
		return NULL;
	}
	/* Copy the prefix, path, and newline char (in this order). */
	strscpy(strings[i], IMAGE_PATH_PREFIX_POST_H, strlen(IMAGE_PATH_PREFIX_POST_H));
	strscpy(strings[i] + strlen(IMAGE_PATH_PREFIX_POST_H), post->image_path, strlen(post->image_path));
	strscpy(strings[i] + strlen(IMAGE_PATH_PREFIX_POST_H) + strlen(post->image_path), LINE_DELIMETER_POST_H, strlen(NULL_CHAR_POST_H));
	strscpy(strings[i] + strlen(IMAGE_PATH_PREFIX_POST_H) + strlen(post->image_path) + 1L, NULL_CHAR_POST_H, strlen(NULL_CHAR_POST_H));

	/* Add the post's text. */
	strings[++i] = (char*) calloc(post->text_length + 1L, sizeof(char));
	if (!strings[i]) {
		for (j = 0L; j < i; j++)
			free(strings[i]);
		free(strings);
		return NULL;
	}
	/* Add the text. */
	strscpy(strings[i], post->text, post->text_length);

	/* Compact the strings array into one string. */
	buffer = strcatarr(strings, NULL, 3L + post->referenced_posts_length + post->referencing_posts_length);

	/* Free the strings array. */
	for (j = 0L; j <= i; j++)
		free(strings[j]);
	free(strings);

	/* Return the resultanlt buffer. */
	return buffer;
}

/* This function deserializes the input and generates a struct thereof. */
/* Modifies the given buffer!!! */
static inline post_t *deserialize_post(char *text) {
	/* Sanity check. */
	assert(text);

	/* Local variables. */
	char *buffer, *tmp_buffer, *ptr, *save_ptr, *text_end, **end_ptr;
	linked_list_t *list;
	node_t *current;
	size_t i, j, referenced_posts_length, referencing_posts_length, text_length, elements, length;
	post_t *post;

	/* Get the text's length. */
	text_length = strlen(text);
	text_end = text + text_length + 1L;

	/* Initialize the linked list. */
	list = initialize_single_linked_list();
	if (!list)
		return NULL;

	/* Start tokenizing. */
	/* Get the ID. */
	ptr = strchr(text, POST_REFERENCED_ID_START_DELIMETER_POST_H);
	if (!ptr) {
		free(list);
		return NULL;
	}
	/* Put null char at the delimeter. */
	*ptr = NULL_CHAR_POST_H;
	/* Grab the ID. */
	length = strlen(text);
	tmp_buffer = (char*) calloc(++length, sizeof(char));	/* Add extra space for null terminator. */
	if (!tmp_buffer) {
		free(list);
		return NULL;
	}
	strscpy(tmp_buffer, text, length);
	/* Move the pointer ahead. */
	ptr++;
	/* Add the ID to the list, but first check that the pointers are valid still. */
	if(!ptr || ptr == text_end || !append_node_data_single_linked_list(list, tmp_buffer)) {
		free(tmp_buffer);
		free(list);
		return NULL;
	}
	/* Set the number of list elements. */
	elements = 1L;

	/* Get the referenced IDs. */
	/* Set the count to zero (initially). */
	referenced_posts_length = 0L;
	/* Save the new starting position. */
	save_ptr = ptr;
	/* Tokenize. */
	ptr = strchr(text, POST_REFERENCED_ID_DELIMETER_POST_H);
	while (ptr) {
		/* Copy the current ID. */
		/* Put null char at the delimeter. */
		*ptr = NULL_CHAR_POST_H;
		/* Grab the ID. */
		length = strlen(save_ptr);
		tmp_buffer = (char*) calloc(++length, sizeof(char));	/* Add extra space for null terminator. */
		if (!tmp_buffer) {
			/* Free the list. */
			clear_list_data_single_linked_list(list, free);
			free(list);
			return NULL;
		}
		strscpy(tmp_buffer, saveptr, length);
		/* Move the pointer ahead. */
		ptr++;
		/* Add the ID to the list, but first check that the pointers are valid still. */
		if (!ptr || ptr == text_end || !append_node_data_single_linked_list(list, tmp_buffer)){
			free(tmp_buffer);
			clear_list_data_single_linked_list(list, free);
			free(list);
			return NULL;
		}
		/* Increase referenced posts count. */
		referenced_posts_length++;
		/* Save the new starting position. */
		save_ptr = ptr;
		/* Tokenize. */
		ptr = strchr(save_ptr, POST_REFERENCED_ID_DELIMETER_POST_H);
	}
	/* Deal with the last referenced post ID. */
	ptr = strchr(save_ptr, POST_REFERENCING_ID_START_DELIMETER_POST_H);
	/* Check if it exists. */
	if (!ptr || ptr == text_end) {
		clear_list_data_single_linked_list(list, free);
		free(list);
		return NULL;
	}
	/* Put null char at the delimeter. */
	*ptr = NULL_CHAR_POST_H;
	/* Grab the ID. */
	length = strlen(save_ptr);
	tmp_buffer = (char*) calloc(++length, sizeof(char));	/* Add extra space for null terminator. */
	if (!tmp_buffer) {
		/* Abort. */
		clear_list_data_single_linked_list(list, free);
		free(list);
		return NULL;
	}
	/* Copy. */
	strscpy(tmp_buffer, save_ptr, length);
	/* Add the ID to the list. */
	if (!append_node_data_single_linked_list(list, tmp_buffer)) {
		free(tmp_buffer);
		clear_list_data_single_linked_list(list, free);
		free(list);
		return NULL;
	}
	/* Increase referencing posts count. */
	referenced_posts_length++;

	/* Get the referencing IDs. */
	/* Set the count to zero (initially). */
	referencing_posts_length = 0L;
	/* Save the new starting position. */
	save_ptr = ptr;
	/* Tokenize. */
	ptr = strchr(text, POST_REFERENCING_ID_DELIMETER_POST_H);
	while (ptr) {
		/* Copy the current ID. */
		/* Put null char at the delimeter. */
		*ptr = NULL_CHAR_POST_H;
		/* Grab the ID. */
		length = strlen(save_ptr);
		tmp_buffer = (char*) calloc(++length, sizeof(char));	/* Add extra space for null terminator. */
		if (!tmp_buffer) {
			/* Free the list. */
			clear_list_data_single_linked_list(list, free);
			free(list);
			return NULL;
		}
		strscpy(tmp_buffer, saveptr, length);
		/* Move the pointer ahead. */
		ptr++;
		/* Add the ID to the list, but first check that the pointers are valid still. */
		if (!ptr || ptr == text_end || !append_node_data_single_linked_list(list, tmp_buffer)){
			free(tmp_buffer);
			clear_list_data_single_linked_list(list, free);
			free(list);
			return NULL;
		}
		/* Increase referencing posts count. */
		referencing_posts_length++;
		/* Save the new starting position. */
		save_ptr = ptr;
		/* Tokenize. */
		ptr = strchr(save_ptr, POST_REFERENCING_ID_DELIMETER_POST_H);
	}
	/* Deal with the last referencing post ID. */
	ptr = strchr(save_ptr, LINE_DELIMETER_POST_H);
	/* Check if it exists. */
	if (!ptr || ptr == text_end) {
		clear_list_data_single_linked_list(list, free);
		free(list);
		return NULL;
	}
	/* Put null char at the delimeter. */
	*ptr = NULL_CHAR_POST_H;
	/* Grab the ID. */
	length = strlen(save_ptr);
	tmp_buffer = (char*) calloc(++length, sizeof(char));	/* Add extra space for null terminator. */
	if (!tmp_buffer) {
		/* Abort. */
		clear_list_data_single_linked_list(list, free);
		free(list);
		return NULL;
	}
	/* Copy. */
	strscpy(tmp_buffer, save_ptr, length);
	/* Add the ID to the list. */
	if (!append_node_data_single_linked_list(list, tmp_buffer)) {
		free(tmp_buffer);
		clear_list_data_single_linked_list(list, free);
		free(list);
		return NULL;
	}
	/* Increase referenced posts count. */
	referencing_posts_length++;

	/* Update the total number of elements. */
	elements += referenced_posts_length + referencing_posts_length;

	/* Update the saveptr. */
	save_ptr = ptr;

	/* Grab the image path. */
	ptr = strchr(save_ptr, LINE_DELIMETER_POST_H);
	/* Seek ahead to the end of the prefix. */
	ptr += strlen(IMAGE_PATH_PREFIX_POST_H) - 1L;
	/* Check if it exists. */
	if (!ptr || ptr == text_end) {
		clear_list_data_single_linked_list(list, free);
		free(list);
		return NULL;
	}
	/* Put null char at the delimeter. */
	*ptr = NULL_CHAR_POST_H;
	/* Move the save pointer. */
	save_ptr = ptr;
	/* Move to the end of the string. */
	ptr = strchr(save_ptr, LINE_DELIMETER_POST_H);
	/* Put null char at the delimeter. */
	*ptr = NULL_CHAR_POST_H;
	/* Check if it exists. */
	if (!ptr || ptr == text_end) {
		clear_list_data_single_linked_list(list, free);
		free(list);
		return NULL;
	}
	/* Copy the image path. */
	length = strlen(save_ptr);
	tmp_buffer = (char*) calloc(++length, sizeof(char));	/* Add extra space for null terminator. */
	if (!tmp_buffer) {
		/* Abort. */
		clear_list_data_single_linked_list(list, free);
		free(list);
		return NULL;
	}
	/* Copy*/
	strscpy(tmp_buffer, save_ptr, length);
	/* Add the image to the list. */
	if (!append_node_data_single_linked_list(list, tmp_buffer)) {
		free(tmp_buffer);
		clear_list_data_single_linked_list(list, free);
		free(list);
		return NULL;
	}

	/* Move to the next line and copy everything to the end. */
	save_ptr = ++ptr;
	ptr = strchr(save_ptr, NULL_CHAR_POST_H);
	length = strlen(save_ptr);
	tmp_buffer = (char*) calloc(++length, sizeof(char));	/* Add extra space for null terminator. */
	if (!tmp_buffer) {
		/* Abort. */
		clear_list_data_single_linked_list(list, free);
		free(list);
		return NULL;
	}
	/* Copy. */
	strscpy(tmp_buffer, save_ptr, length);
	/* Add the text to the list. */
	if (!append_node_data_single_linked_list(list, tmp_buffer)) {
		free(tmp_buffer);
		clear_list_data_single_linked_list(list, free);
		free(list);
		return NULL;
	}

	/* Traverse the list and build the struct. */
	/* Attempt to allocate the struct. */
	post = (post_t*) malloc(sizeof(post_t));
	if (!post) {
		/* Abort. */
		clear_list_data_single_linked_list(list, free);
		free(list);
		return NULL;
	}
	/* Get the OP ID. */
	current = list->head;
	post->id = strtoull(((char*) current->data), end_ptr, NUMBER_PARSING_BASE_POST_H);
	if (!end_ptr || **end_ptr) {
		/* Abort. */
		clear_list_data_single_linked_list(list, free);
		free(list);
		free(post);
		return NULL;
	}
	/* Deal with the referenced IDs. */
	post->referenced_posts_length = referenced_posts_length;
	/* Attempt to allocate the array. */
	post->referenced_posts = (unsigned long long *) calloc(posts->referenced_posts_length, sizeof(unsigned long long));
	if (!post->referenced_posts) {
		/* Abort. */
		clear_list_data_single_linked_list(list, free);
		free(list);
		free(post);
		return NULL;
	}
	/* Loop over the values and add them to the array. */
	for (i = 0L; i < post->referenced_posts_length; i++) {
		/* Move the current pointer ahead. */
		current = current->next;
		assert(current);	/* For debug, shouldn't be needed in prodcution. */
		/* Attempt to parse the current ID. */
		post->referenced_posts[i] = strtoull(((char*) current->data), end_ptr, NUMBER_PARSING_BASE_POST_H);
		if (!end_ptr || **end_ptr) {
			/* Abort. */
			clear_list_data_single_linked_list(list, free);
			free(list);
			free(post->referenced_posts);
			free(post);
			return NULL;
		}
	}
	/* Deal with the referencing IDs. */
	post->referencing_posts_length = referencing_posts_length;
	/* Attempt to allocate the array. */
	post->referencing_posts = (unsigned long long *) calloc(posts->referencing_posts_length, sizeof(unsigned long long));
	if (!post->referencing_posts) {
		/* Abort. */
		clear_list_data_single_linked_list(list, free);
		free(list);
		free(post->referenced_posts);
		free(post);
		return NULL;
	}
	/* Loop over the values and add them to the array. */
	for (i = 0L; i < post->referencing_posts_length; i++) {
		/* Move the current pointer ahead. */
		current = current->next;
		assert(current);	/* For debug, shouldn't be needed in prodcution. */
		/* Attempt to parse the current ID. */
		post->referencing_posts[i] = strtoull(((char*) current->data), end_ptr, NUMBER_PARSING_BASE_POST_H);
		if (!end_ptr || **end_ptr) {
			/* Abort. */
			clear_list_data_single_linked_list(list, free);
			free(list);
			free(post->referenced_posts);
			free(post->referencing_posts);
			free(post);
			return NULL;
		}
	}
	/* Image path. */
	current = current->next;
	assert(current);	/* For debug, shouldn't be needed in production. */
	/* Copy the image path. */
	length = strlen((char*) current->data);
	tmp_buffer = (char*) calloc(++length, sizeof(char));	/* Added space for null-terminator. */
	if (!tmp_buffer) {
		/* Abort. */
		clear_list_data_single_linked_list(list, free);
		free(list);
		free(post->referenced_posts);
		free(post->referencing_posts);
		free(post);
		return NULL;
	}
	strscpy(tmp_buffer, ((char*) current->data), length);
	post->image_path = tmp_buffer;
	/* Post text. */
	current = current->next;
	assert(current && !(current->next));	/* For debug, shouldn't be needed in production. */
	/* Copy the image path. */
	length = strlen((char*) current->data);
	tmp_buffer = (char*) calloc(++length, sizeof(char));	/* Added space for null-terminator. */
	if (!tmp_buffer) {
		/* Abort. */
		clear_list_data_single_linked_list(list, free);
		free(list);
		free(post->referenced_posts);
		free(post->referencing_posts);
		free(post->image_path);
		free(post);
		return NULL;
	}
	strscpy(tmp_buffer, ((char*) current->data), length);
	post->text = tmp_buffer;

	/* Cleanup. */
	clear_list_data_single_linked_list(list, free);
	free(list);

	/* Return the struct. */
	return post;
}
