/*	lemon_board Copyright (C) 2022 Wael Karram.
*
*	This program is free software: you can redistribute it and/or modify
*	it under the terms of the GNU Affero General Public License as published by
*	the Free Software Foundation, version 3 of the License.
*
*	This program is distributed in the hope that it will be useful,
*	but WITHOUT ANY WARRANTY; without even the implied warranty of
*	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*	GNU Affero General Public License for more details.
*
*	You should have received a copy of the GNU Affero General Public License
*	along with this program.  If not, see <https://www.gnu.org/licenses/>.
*/

/*
* Implementation notes:
* The serialized data is structured as follows:
* OP_ID followed by a start delimeter for referenced IDs, which in turn is followed by the referenced IDs delimeter after each such ID.
* At the end there is a new start delimeter for the referencing IDs, similarly they are separated by their own delimeters and then a new line at the end.
* The image path follows the IDs, starting with a prefix and ending with a new line.
* At the end is the post's text, which should be followed by EOF.
*/

/* Include guard. */
#ifndef POST_STRUCT
#define POST_STRUCT

/* Include stdlib for string to unsigned long long conversion. */
#include <stdlib.h>
#define _GNU_SOURCE /* Using GCC extensions for strtoull. */

/* Include utils for strings. */
#include "../util/string_util.h"

/* Include strcatarr and strapp from C_lib. */
#include "../C_lib/strings/strcatarr.h"
#include "../C_lib/strings/strapp.h"

/* Include linked lists (single-linked). */
#include "../C_lib/linked_lists/single_linked_list.h"

/* Define constants used for serializing and deserializing the data. */
#define POST_REFERENCED_ID_START_DELIMETER_POST_H	'-'		/* Post referenced IDs after the OP. */
#define POST_REFERENCED_ID_DELIMETER_POST_H		','		/* Delimeter for referenced post IDs after the OP. */
#define POST_REFERENCING_ID_START_DELIMETER_POST_H	'+'		/* Starting delimeter for referencing post IDs after referenced IDs. */
#define POST_REFERENCING_ID_DELIMETER_POST_H		'='		/* Delimeter for referencing post IDs after referenced IDs. */
#define LINE_DELIMETER_POST_H				'\n'		/* Delimeter between lines. */
#define NULL_CHAR_POST_H				'\0'		/* Null char to indicate end of string. */
#define IMAGE_PATH_PREFIX_LENGTH_POST_H			12L		/* Define the length of the image path prefix. */
#define NUMBER_PARSING_BASE_POST_H			10		/* Define the base used to parse number strings. */

/* String literal constants. */
static const char[] IMAGE_PATH_PREFIX_POST_H =		"IMAGE_PATH: ";	/* Prefix to the image path. */
static const char[] NEWLINE_STRING_POST_H =		"\n";		/* String constant of a newline. */

/* Define data structures. */
/* This struct is used to store a single post. */
typedef struct Post {
	/* Post ID. */
	unsigned long long id;
	/* Referenced posts array length. */
	size_t referenced_posts_length;
	/* Referencing posts array length. */
	size_t referencing_posts_length;
	/* Referenced posts array. */
	unsigned long long *referenced_posts;
	/* Referencing posts array. */
	unsigned long long *referencing_posts;
	/* Post text. */
	char *text;
	/* Length of text. */
	size_t text_length;
	/* Image path. */
	char *image_path;	/* Not much of a point in putting a length as this should be somewhat short. */
} post_t;

/* Define a serialization function. */
static inline char *serialize_post(post_t *post);
/* Define a generation/deserialization function, modifies the input buffer!! */
static inline post_t *deserialize_post(char *text);

#endif	/* POST_STRUCT */
