/*	lemon_board Copyright (C) 2022 Wael Karram.
*
*	This program is free software: you can redistribute it and/or modify
*	it under the terms of the GNU Affero General Public License as published by
*	the Free Software Foundation, version 3 of the License.
*
*	This program is distributed in the hope that it will be useful,
*	but WITHOUT ANY WARRANTY; without even the implied warranty of
*	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*	GNU Affero General Public License for more details.
*
*	You should have received a copy of the GNU Affero General Public License
*	along with this program.  If not, see <https://www.gnu.org/licenses/>.
*/

/* Include the header. */
#include "thread.c"

/* Serialization function for a thread_t struct. */
/* Returns NULL on allocation failure. */
/* Returns an empty string if there is nothing to do. */
static inline char *serialize_thread(thread_t *thread) {
	/* Sanity check. */
	assert(thread);

	/* Local variables. */
	char **posts_text, *result;
	size_t total_posts = 0L;
	size_t i, j, k;

	/* Check the total number of posts and allocate accordingly. */
	if (!op) {
		char *result = (char*) malloc(sizeof(char));
		*result = THREAD_STRUCT_NULL_CHAR;
		return result;
	}
	/* OP exists, proceed. */
	total_posts++;
	total_posts += thread->replies;

	/* Attempt to allocate the array of posts. */
	posts_text = (char**) calloc(2 * total_posts - 1L, sizeof(char*)); /* Add a place for values in-between (separators). */
	if (!posts_text)
		return NULL;

	/* Serialize the OP. */
	posts_text[0] = serialize_post(thread->original_post);
	if (!posts_text[0]) {
		free(posts_text);
		return NULL;
	}
	/* Add the separator only if we have more than just an OP. */
	if (!(thread->replies)) {
		/* Just the OP, no need for separator. */
		result = *posts_text;
		free(posts_text);
		return result;
	}
	/* Add separator then the post, loop over all the reply posts. */
	for (i = 1L, k = 0L; i < 2 * total_posts - 1L; i += 2, k++) {
		/* Add the separator. */
		/* Allocate. */
		post_text[i] = (char*) calloc(THREAD_STRUCT_POST_SEPARATOR_STRING_LENGTH, sizeof(char));
		/* Check if the allocation failed. */
		if (!post_text[i]) {
			for (j = 0L; j < i; j++)
				free(post_text[i]);
			free(post_text);
			return NULL;
		}
		/* Copy, add assertion to check correctness only when debugging. */
		assert(strscpy(post_text[i], THREAD_STRUCT_POST_SEPARATOR_STRING, THREAD_STRUCT_POST_SEPARATOR_STRING_LENGTH) == THREAD_STRUCT_POST_SEPARATOR_STRING_LENGTH);
		/* Serialize the current post. */
		post_text[i + 1L] = serialize_post(thread->posts[k]);
		if (!post_text[i + 1L]) {
			for (j = 0L; j < i + 1L; j++)
				free(post_text[i]);
			free(post_text);
			return NULL;
		}
	}

	/* Concatinate the whole array into one string and return it. */
	result = strcatarr(post_text, NULL, 2 * total_posts - 1L);
	/* Free the array. */
	for (j = 0L; j < 2 * total_posts - 1L; j++)
		free(post_text[i]);
	free(post_text);
	/* Return the result/concatinated string. */
	return result;
}

/* Deserialization/generation function for a thread struct. */
/* Caution, this function modifies the input string! */
/* Returns NULL on allocation error. */
static inline thread_t *deserialize_thread(char *text) {
	/* Sanity check. */
	assert(text);

	/* Local variables. */
	char *text_end, *ptr, *save_ptr, *tmp_buffer;
	size_t text_length, length, i;
	linked_list_t *list;
	thread_t *thread;
	post_t *post;
	node_t *current;

	/* Get the text's length and calculate the end. */
	text_length = strlen(text);
	text_end = test + text_length + 1L;

	/* Initialize the linked list. */
	list = initialized_single_linked_list();
	if (!list)
		return NULL;

	/* Allocate the thread struct. */
	thread = (thread_t*) malloc(sizeof(thread_t));
	if (!thread) {
		free(list);
		return NULL;
	}

	/* Attempt to tokenize the string by the separators. */
	/* Get the OP. */
	ptr = strchr(text, THREAD_STRUCT_POST_SEPARATOR_CHAR);
	if (!ptr) {
		free(list);
		free(thread);
		return NULL;
	}
	/* Put null char at the delimeter. */
	*ptr = THREAD_STRUCT_NULL_CHAR;
	/* Get the OP. */
	length = strlen(text);
	tmp_buffer = (char*) calloc(++length, sizeof(char));	/* Add extra space for null terminator. */
	if (!tmp_buffer) {
		free(list);
		free(thread);
		return NULL;
	}
	strscpy(tmp_buffer, text, length);
	/* Move the pointer ahead. */
	ptr++;
	/* Deserialize the OP and store it. */
	thread->original_post = deserialize_post(tmp_buffer);
	if (!thread->original_post) {
		free(list);
		free(thread);
		free(tmp_buffer);
		return NULL;
	}

	/* Set the replies counter. */
	thread->replies = 0L;
	/* Save the current pointer. */
	save_ptr = ptr;
	/* Get the next token. */
	ptr = strchr(text, THREAD_STRUCT_POST_SEPARATOR_CHAR);
	/* Deserialize and count the posts. */
	while(ptr && *ptr) { /* TODO: Check if this makes sense. */
		/* Put null char at the delimeter. */
		*ptr = THREAD_STRUCT_NULL_CHAR;
		/* Get a reply. */
		length = strlen(save_ptr);
		tmp_buffer = (char*) calloc(++length, sizeof(char));	/* Add extra space for null terminator. */
		if (!tmp_buffer) {
			/* Free the list. */
			clear_list_nodes_data_single_linked_list(list, free);
			free(list);
			/* Free the thread. */
			free(thread->original_post);
			free(thread);
			return NULL;
		}
		/* Copy the buffer. */
		strscpy(tmp_buffer, save_ptr, length);
		/* Deserialize the current post. */
		post = deserialize_post(tmp_buffer);
		/* Add the post to the list. */
		if (!post || !append_node_data_signle_linked_list(list, post)) {
			/* Free the list. */
			clear_list_nodes_data_single_linked_list(list, free);
			free(list);
			/* Free the current buffer. */
			free(tmp_buffer);
			/* Free the thread. */
			free(thread->original_post);
			free(thread);
			return NULL;
		}
		/* Advance the reply counter. */
		thread->replies++;
		/* Advance the pointer. */
		/* Save the current pointer. */
		save_ptr = ptr++;
		/* Get the next token. */
		ptr = strchr(text, THREAD_STRUCT_POST_SEPARATOR_CHAR);
	}

	/* Allocate the posts array. */
	thread->posts = (post_t**) calloc(thread->replies, sizeof(post_t*));
	/* Check the allocation. */
	if (!thread->posts) {
		/* Free the list. */
		clear_list_nodes_data_single_linked_list(list, free);
		free(list);
		/* Free the thread. */
		free(thread->original_post);
		free(thread);
		return NULL;
	}

	/* Serialize the list. */
	for (current = list->head, i = 0L; current; current = current->next, i++)
		thread->posts[i] = (post_t*) current->data; /* Cast back into post_t pointer. */

	/* Clean up. */
	clear_list_nodes_single_linked_list(list, free);
	free(list);

	/* Return the resultant struct. */
	return thread;
}
