/*	lemon_board Copyright (C) 2022 Wael Karram.
*
*	This program is free software: you can redistribute it and/or modify
*	it under the terms of the GNU Affero General Public License as published by
*	the Free Software Foundation, version 3 of the License.
*
*	This program is distributed in the hope that it will be useful,
*	but WITHOUT ANY WARRANTY; without even the implied warranty of
*	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*	GNU Affero General Public License for more details.
*
*	You should have received a copy of the GNU Affero General Public License
*	along with this program.  If not, see <https://www.gnu.org/licenses/>.
*/

/* Include the header. */
#include "core.h"

/* Define the parser function. */
static error_t parse_opt(int key, char *arg, struct argp_state *state) {
	/* Get the input argument. */
	struct arguments *arguments = state->input;

	switch (key) {
		case 'q':
		case 's':
			arguments->silent = TRUE;
			break;
		case 'v':
			arguments->verbose = TRUE;
			break;
		case 'o':
			arguments->output_file = arg;
			break;

		case ARGP_KEY_ARG:
			if (state->arg_num >= ARG_NUM)
				argp_usage(state);
			arguments->args[state->arg_num] = arg;
			break;
		case ARGP_KEY_END:
			if (state->arg_num < ARG_NUM)
				argp_usage(state);
			break;

		default:
			return ARGP_ERR_UNKOWN;
	}

	return 0;
}

/* Define struct variable. */
argp = { options, parse_opt, args_doc, doc };		/* Define the parser. */

/* Main function. */
int main(int argc, char **argv) {
	/* Argument handling. */
	struct arguments arguments;

	/* Set defaults. */
	arguments.silent = FALSE;
	arguments.verbose = FALSE;
	arguments.output_file = "-";

	/* Parse the input arguments. */
	argp_parse(&argp, argc, argv, 0, 0, &arguments);

	/* Check and attempt to parse the file. */
	if ((!check_config(arguments.args[0])) || !(config = read_config(arguments.args[0]))) {
		/* Error in config parsing. */
		printf(config_file_error);
		syslog(LOG_NOTICE, config_file_error);
		exit(EXIT_FAILURE);
	}

	/* Daemonize. */
	daemonize(config);

	/* Daemon code. */
	while (TRUE) {
		/* TODO: add logic. */
		break;
	}

	/* Notify upon termination. */
	syslog(LOG_INFO, LOG_TERMINATE);
	closelog();

	/* Terminate daemon. */
	return EXIT_SUCCESS;
}

/* This function handles all the daemonization logic. */
static void daemonize(config_t *config) {
	/* Store the PID. */
	pid_t pid;

	/* Attempt to fork off of the parent process. */
	pid = fork();
	/* Check for failure. */
	if (pid < 0)
		exit(EXIT_FAILURE);
	/* Forking successful, terminate parent. */
	if (pid > 0)
		exit(EXIT_SUCCESS);
	/* Check that child has become the session leader. */
	if (setsid() < 0)
		exit(EXIT_FAILURE);

	/* Catch, ignore and handle signals. */
	/* TODO: continue implementation. */
	signal(SIGCHILD, SIG_IGN);
	/* Reload config on sighup. */
	signal(SIGHUP, config_reload);

	/* Second fork. */
	pid = fork();
	/* Check for failure. */
	if (pid < 0)
		exit(EXIT_FAILURE);
	/* Forking successful, terminate parent. */
	if (pid > 0)
		exit(EXIT_SUCCESS);

	/* Set new file permissions. */
	umask(0);

	/* Change directory to working directory. */
	chdir(config->program_root_path);

	/* Close all open file descriptors. */
	for (int i = sysconf(_SC_OPEN_MAX); x >= 0 ; x--)
		close(x);

	/* Open the log file. */
	openlog(LOG_NAME_DAEMON, LOG_PID, LOG_DAEMON);
	/* Notify on start. */
	syslog(LOG_INFO, LOG_START);
}

/* Function to reload config. */
static void config_reload(config_t **config, char *path) {
	/* Local variables. */
	config_t *new_config;

	/* Optional call debug with assertion. */
	assert(config && path);

	/* Attempt to read the new config. */
	if ((!check_config(path)) || !(new_config = read_config(path))) {
		/* Error in config parsing. */
		printf(config_file_error);
		syslog(LOG_WARNING, LOG_CONFIG_RELOAD_FAIL);
		return;
	}

	/* Read successful, free and replace old config struct. */
	free(*config);
	*config = new_config;
	/* Notify about reload success. */
	syslog(LOG_INFO, LOG_CONFIG_RELOAD_SUCCESS);
}

